// devicelist.h

#ifndef _DEVICELIST_H_
#define _DEVICELIST_H_

#include <genesis/devices.h>
#include <genesis/genesis.h>

class data_connection;

//----------------------------- nodes ---------------------

struct readclient
{
    readclient();

    data_connection* conn_;
    int lock_mask;          // which channels are read-locked by this client
};


struct writeclient
{
    writeclient();

    data_connection* conn_;
    int lock_mask;          // which channels are write-locked by this client
};


class devicenode
{
public:
    devicenode(data_connection* conn, const char* owner, int dev_type, 
               const char* desc, int nr_elements);
    ~devicenode();

    int type() { return type_; }
    data_connection* connection() { return conn_; }
    const char* cookie() { return cookie_; }
    const char* owner() { return owner_; }

    void create_cookie();

    int add_readclient(data_connection* gclient);
    int add_writeclient(data_connection* gclient);
    int add_readwriteclient(data_connection* gclient);

    int nr_readclients() const { return readcount_; }
    readclient* nth_readclient(int n) const;
        // returns NULL if not nth client

    int nr_writeclients() const { return writecount_; }
    writeclient* nth_writeclient(int n) const;
        // returns NULL if not nth client

    void remove_client(const data_connection* gclient);

    int has_write_access(const data_connection* gclient) const;

    int writeslot_claimed();
        // returns 1 if some writeslot is claimed, else -1
    int readslot_claimed();
        // return 1 if some readslot is claimed, else -1

protected:
    int writeslot_available();
        // returns index of slot if available, else -1
    int readslot_available();
        // returns index of slot if available, else -1
    void remove_readclient(const data_connection* gclient);
    void remove_writeclient(const data_connection* gclient);

    data_connection* conn_;
    int nr_elements_;
    int type_;
    char description_[max_description_size];
    readclient** readclients_;
    int readcount_;
    writeclient** writeclients_;
    int writecount_;
    char* owner_;               // user in user@host format
    char cookie_[cookie_size + 1];
    int demo;			// for demo xxxxxx cookies JvO
};

//------------------------------ list --------------------------

class devicelist
{
public:
    devicelist(int nr_elements = 64, int max_clients_per_device = 16);
    ~devicelist();

    int id_ok(int device_id) const;
        // returns 1 of device_id is an id of an existing device, 0 if not

    int register_device(data_connection* device, int dev_type, 
                        const char* desc, char* cookie, const char* username);
        // returns device-id, -1 if failed
    void unregister_device(int device_id);
    void unregister_device(const data_connection* gdevice);

    data_connection* device_conn(int device_id) const;
        // returns NULL if no such device
    int device_id(const data_connection* gdevice) const;
        // returns -1 if no such device

    int request_device(data_connection* gclient, int dev_type,
                       const char* cookie, const char* username);
        // returns device-if of granted device, -1 if failed
    void release_device(const data_connection* gclient, int device);

    int has_device(int device_id) const;
    int has_device(const data_connection* gdevice) const;

    int nr_readclients(int device_id) const;
        // returns -1 if device does not exist
    readclient* nth_readclient(int device_id, int n) const;

    int nr_writeclients(int device_id) const;
        // returns -1 if device does not exist
    writeclient* nth_writeclient(int device_id, int n) const;

    int has_write_access(data_connection* gclient, int device_id) const;

    void remove_client(data_connection* gclient);

protected:
    devicenode** list_;
    int nr_elements_;
    int max_clients_;
};

//------------------------- iterators ---------------------------

class devicelist_iterator
{
public:
    devicelist_iterator(devicelist* list, int device_id);

    virtual int ok() { return ok_; }
    void first();
    void next();
    virtual int done() = 0;
    
protected:
    int ok_;
    devicelist* list_;
    int current_;
    int device_id_;
};


class readlist_iterator: public devicelist_iterator
{
public:
    readlist_iterator(devicelist* list, int device_id);

    int done();
    readclient* item() const;
};


class writelist_iterator: public devicelist_iterator
{
public:
    writelist_iterator(devicelist* list, int device_id);

    int done();
    writeclient* item() const;
};


#endif _DEVICELIST_H_
