// midiserver.cc

#include "midiserver.h"
#include "devicelist.h"
#include "eden.h"

#include <cs/tcp/server.h>
#include <cs/dataconn.h>
#include <cs/samlib.h>
#include <genesis/msgparser.h>
#include <genesis/midiparser.h>
#include <genesis/msgtypes.h>

#include <errno.h>
#include <signal.h>
#include <stdio.h>
#include <stream.h>
#include <string.h>


const int device_mask = 1 << 31;
int user_interrupt = 0;


void inthandler(int /* signr */)
{
    user_interrupt = 1;
    cout << "^C pressed" << endl;
}


static void printmsg(const char* msg, int msglen)
{
    int i;

    for (i = 0; i < msglen; i++)
    {
        printf("%x ", (unsigned int) msg[i]);
    }
    printf("\n");
}


midiserver::midiserver(int portnr)
{
    msgparser_ = new msgparser();
    midiparser_ = new midiparser();
    if (portnr > 0)
        server_ = new tcp_server(portnr);
    else
        server_ = new tcp_server();
    devlist_ = new devicelist();
}


midiserver::~midiserver()
{
    if (eden_trace)
        cout << "Shutting down server" << endl;

    delete midiparser_;
    delete msgparser_;
    delete server_;
    delete devlist_;
}


int midiserver::main(int /* argc */, char* /* argv */[])
{
    char buffer[1024];
    int nrbytes = 0;

    cerr << "Server is at port " << server_ -> portnr() << endl;

    signal(SIGINT, inthandler);

    while(!user_interrupt)
    {
        if (!server_ -> connected())
            return -1;

        data_connection* conn = server_ -> select();
 
        if (conn != NULL)
        {
            nrbytes = server_ -> readmsg(conn, buffer, sizeof(buffer));
            if (eden_trace)
                cout << "Received " << nrbytes << " bytes" << endl;
     
            if (nrbytes > 0)
                process(conn, buffer, nrbytes);
            else
            {
                if (eden_trace)
                    cout << "Probably disconnecting client or device" << endl;
                if ((conn == NULL) || (!conn -> connected()))
                {
                    if (eden_trace)
                        cout << "yep, something not connected anymore" << endl;
                // now find out what we have to remove, a client or device
                    if (devlist_ -> has_device(conn))
                        devlist_ -> unregister_device(conn);
                    else
                        devlist_ -> remove_client(conn);
                    server_ -> close(conn);
                }
                else
                {
                    if (eden_trace)
                        cout << "    nope, other error" << endl;
                }
            }
        }
        else
        {
            if (eden_trace)
                cout << "select returns NULL" << endl;
        }
    }
 
    return 0;
}


void midiserver::process(data_connection* conn, const char* msg, int msglen)
{
    switch (msgparser_ -> type(msg))
    {
        case msg_request_dev:
            request_device(conn, msg, msglen);
            break;
        case msg_release_dev:
            release_device(conn, msg, msglen);
            break;
        case msg_register_dev:
            register_device(conn, msg, msglen);
            break;
        case msg_unregister_dev:
            unregister_device(conn, msg, msglen);
            break;
        case msg_midi_multicast:
            midi_multicast(conn, msg, msglen);
            break;
        case msg_midi_out:
            midi(conn, msg, msglen);
            break;
        default:
            if (eden_trace)
                cout << "got message type " << msgparser_ -> type(msg) << endl;
    }
}


void midiserver::request_device(data_connection* conn, const char* msg, 
                                int /* msglen */)
{
    char msgout[128];
    int len;
    char cookie[cookie_size+1];
    char username[64];
    int dev_type = (int) msgparser_ -> device_type(msg);
    msgparser_ -> get_cookie(msg, cookie);

    if (!msgparser_ -> parse_request_dev(msg, dev_type, cookie, username))
    {
        warn("midiserver::request_device : invalid request message");
        return;
    }

    int device_id = devlist_ -> request_device(conn, dev_type, cookie,
                                               username);
    
    if (eden_trace)
    {
        cout << "    request of device type " << dev_type;
        cout << " by user " << username << endl;
    }
    if (device_id >= 0)
    {
        len = msgparser_ -> create_grant_dev(msgout, device_id);
        server_ -> writemsg(conn, msgout, len);
        if (eden_trace)
            cout << "        granted (dev-id " << device_id << ")" << endl;
    }
    else
    {
        len = msgparser_ -> create_reject_dev(msgout);
        server_ -> writemsg(conn, msgout, len);
        if (eden_trace)
            cout << "        rejected" << endl;
    }
}


void midiserver::release_device(data_connection* conn, const char* msg, 
                                int /* msglen */)
{
    int dev_id = msgparser_ -> device_id(msg);

    devlist_ -> release_device(conn, dev_id);
    if (eden_trace)
        cout << "    release device " << dev_id << endl;
}


void midiserver::register_device(data_connection* conn, const char* msg, 
                                 int /* msglen */)
{
    char msgout[128];
    char desc[max_description_size];
    char username[64];
    int len;
    int dev_type;

    if (!msgparser_ -> parse_register_dev(msg, dev_type, desc, username))
    {
        warn("midiserver::register_device : invalid register message");
        return;
    }

    char cookie[cookie_size+1] = {'\0'};    // filled in by 
                                            // devlist -> register_device
    int device_id = devlist_ -> register_device(conn, dev_type, desc, cookie,
                                                username);

    if (eden_trace)
    {
        cout << "    register_device \"" << desc << "\" type " << dev_type;
        cout << ", owned by " << username << endl;
    }
    if (device_id >= 0)
    {
    // mark device-id as id handled by a device
        device_id |= device_mask;

        len = msgparser_ -> create_accept_dev(msgout, device_id, cookie);
        server_ -> writemsg(conn, msgout, len);
        if (eden_trace)
            cout << "        accepted (dev " << device_id << ")" << endl;
    }
    else
    {
        len = msgparser_ -> create_reject_dev(msgout);
        server_ -> writemsg(conn, msgout, len);
        if (eden_trace)
            cout << "        rejected" << endl;
    }
}


void midiserver::unregister_device(data_connection* /* conn */, 
                                   const char* msg, int /* msglen */)
{
    int dev_id = msgparser_ -> device_id(msg);

// unmask device-mask
    dev_id &= ~device_mask;
    devlist_ -> unregister_device(dev_id);
    if (eden_trace)
        cout << "    unregister device " << dev_id << endl;
}


void midiserver::midi(data_connection* conn, const char* msg, int msglen)
{
    int dev_id = msgparser_ -> device_id(msg);

    if (eden_trace)
        printf("receiving midi-data from device %d\n", dev_id);
// does this come from a client or a device?
    if (dev_id & device_mask)           // from device
    {
        if (eden_trace)
            printf("about to send midi-data to all connected clients\n");
        // should send midi-message to all clients connected to device (FIX ME)
        readlist_iterator riter(devlist_, dev_id ^ device_mask);

        riter.first();
        while (!riter.done())
        {
            readclient* rc = riter.item();

            if (eden_trace)
                cout << "    sending midi-data from device to client" << endl;

            if (rc != NULL)
                server_ -> writemsg(rc -> conn_, msg, msglen);
            else
                cerr << "midiserver::midi : rc == NULL" << endl;

            riter.next();
        }
    }
    else                                // from client
    {
        data_connection* devconn = devlist_ -> device_conn(dev_id);

        if (devlist_ -> has_write_access(conn, dev_id) && (devconn != NULL))
        {
            server_ -> writemsg(devconn, msg, msglen);
            if (eden_trace)
                cout << "    sending midi-data from client to device" << endl;
        }
        else
        {
        // an error message should be sent back (FIX ME)
            if (eden_trace)
                cout << "    device not available" << endl;
        }
    }
}


void midiserver::midi_multicast(data_connection* conn, const char* msg, 
                                int msglen)
{
    int dev_id = msgparser_ -> device_id(msg);
    int dev_ids[64];
    int nrdevs;
    char mididata[128];
    int midilen;
    char newmsg[128];
    int newmsglen;
    int i;
    int srcdev;

    if (eden_trace)
        printmsg(msg, msglen);

// does this come from a client or a device?
    if (dev_id & device_mask)           // from device
    {
        if (eden_trace)
        {
            cout << "midiserver::midi_multicast : multicast from device, ";
            cout << "ignored" << endl;
        }
        return;
    }
    else                                // from client
    {
        if (!msgparser_ -> parse_midi_multicast(msg, srcdev, dev_ids, nrdevs,
                                                mididata, midilen))
        {
            cerr << "midiserver::midi_multicast : couldn't parse msg" << endl;
            return;
        }
        newmsglen = msgparser_ -> create_midi(newmsg, dev_id, mididata, 
                                              midilen);

        if (eden_trace)
            printmsg(newmsg, newmsglen);
        for (i = 0; i < nrdevs; i++)
        {
            dev_id = dev_ids[i];

            if (eden_trace)
            {
                cout << "trying dev-id " << dev_id << endl;
            }

            data_connection* devconn = devlist_ -> device_conn(dev_id);

            if (devlist_ -> has_write_access(conn, dev_id) && (devconn != NULL))
            {
                server_ -> writemsg(devconn, newmsg, newmsglen);
                if (eden_trace)
                {
                    cout << "    sending midi-data from client to device";
                    cout << endl;
                }
            }
            else
            {
            // an error message should be sent back (FIX ME)
                if (eden_trace)
                {
                    cout << "    device not available" << endl;
                    if (devconn == NULL)
                        cout << "    no device with id " << dev_id << endl;
                    if (!devlist_ -> has_write_access(conn, dev_id))
                    {
                        cout << "    client can not write to dev ";
                        cout << dev_id << endl;
                    }
                }
            }
        }
    }
}
