// hushjam.cc

#include <genesis/jamclient.h>
#include <genesis/midi.h>
#include <cs/samlib.h>

#include <iostream.h>
#include <unistd.h>
#include <stdlib.h>
#include <string.h>

#include <hush/session.h>
#include <hush/event.h>
#include <widgets/button.h>
#include <widgets/canvas.h>
#include <widgets/label.h>
#include <widgets/scale.h>

#include <tk.h>
#include <X11/Xlib.h>


const char* cmd_quit = "Quit";
const char* cmd_keypress = "Keypress";
const char* cmd_keyrelease = "Keyrelease";
const char* cmd_join = "JoinJam";
const char* cmd_leave = "LeaveJam";
const char* cmd_volume = "SetVolume";
const char* cmd_instrument = "SetInstrument";
const char* cmd_octave = "SetOctave";
const char* event_enter_window = "EnterWindow";
const char* event_leave_window = "LeaveWindow";


//------------------------- definition ----------------------------

class c_misc_handler;

class app: public session 
{
public:
    app(int argc, char* argv[]);
    int main();
    void disconnect();
    void connect();
    void keypress(char ch);
    void keyrelease(char ch);
    void set_octave();
    void set_instrument();
    void set_volume();
    void keyboard_press(const char* playstr);   // key on 'piano'
    void keyboard_release(const char* releasestr);
    void enter_window();                        // mouse-cursor leaves window
    void leave_window();                        // mouse-cursor enters window
    void quit();

protected:
    c_misc_handler* misc_handler_;
    jamclient* pclient_;
    int octave_;
    int volume_;
    int instrument_;

    scale* svolume_;
    scale* sinstrument_;
    scale* soctave_;

// obscure struct to remove/restore keyrepeat:
    XKeyboardState orgstate_;
    Display* display_;
};


class c_misc_handler: public handler
{
public:
    c_misc_handler(app* owner);
    int operator()();

protected:
    app* owner_;
};


//------------------------- implementation ----------------------------

//---- c_misc_handler ----

c_misc_handler::c_misc_handler(app* owner)
{
    owner_ = owner;
}


int c_misc_handler::operator()()
{
    //printf("c_misc_handler::operator()\n");
    //for (int i = 0; i < _event -> argc(); i++)
        //printf("    arg %d = %s\n", i, _event -> arg(i));
    //printf("keyevent = %d\n", _event -> keyevent());
    //printf("keycode = %d\n", _event -> keycode());

    if (strcmp(_event -> arg(0), "play") == 0)
    {
        owner_ -> keyboard_press(_event -> arg(1));
        return OK;
    }
    if (strcmp(_event -> arg(0), "release") == 0)
    {
        owner_ -> keyboard_release(_event -> arg(1));
        return OK;
    }

    if (strcmp(_event -> arg(1), cmd_quit) == 0)
        owner_ -> quit();
    else if (strcmp(_event -> arg(1), cmd_leave) == 0)
        owner_ -> disconnect();
    else if (strcmp(_event -> arg(1), cmd_join) == 0)
        owner_ -> connect();
    else if (strncmp(_event -> arg(1), cmd_keypress, 
                     strlen(cmd_keypress)) == 0)
        owner_ -> keypress(_event -> arg(2)[0]);
    else if (strncmp(_event -> arg(1), cmd_keyrelease, 
                     strlen(cmd_keyrelease)) == 0)
        owner_ -> keyrelease(_event -> arg(2)[0]);
    else if (strcmp(_event -> arg(1), cmd_instrument) == 0)
        owner_ -> set_instrument();
    else if (strcmp(_event -> arg(1), cmd_volume) == 0)
        owner_ -> set_volume();
    else if (strcmp(_event -> arg(1), cmd_octave) == 0)
        owner_ -> set_octave();
    else if (strcmp(_event -> arg(1), event_enter_window) == 0)
        owner_ -> enter_window();
    else if (strcmp(_event -> arg(1), event_leave_window) == 0)
        owner_ -> leave_window();
    else
        warn("c_misc_handler::operator() : unknown command string %s",
             _event -> arg(1));

    return OK;
}

//---- app ----

app::app(int argc, char* argv[]) 
   :session(1, argv, "HushJammer") 
{ 
    char cookie[cookie_size+1] = "";
    char incookie[cookie_size+1] = "";
    char host[64] = JAMSERVER_DEFAULT_HOST;
    int port = jamserver_default_port;
    int argindex = 1;

    parse_jamserver_env(host, port);

    while (argindex < argc)
    {
        if (strcmp(argv[argindex], "-trace") == 0)
        {
            cout << "Tracing on" << endl;
            tk -> trace();
            argindex++;
        }
        else if (strcmp(argv[argindex], "-port") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing port" << endl;
                exit(1);
            }
            port = atoi(argv[argindex]);
            cout << "Port set to " << port << endl;
            argindex++;
        }
        else if (strcmp(argv[argindex], "-host") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing hostname" << endl;
                exit(1);
            }
            strcpy(host, argv[argindex]);
            cout << "Host set to " << host << endl;
            argindex++;
        }
        else if (strcmp(argv[argindex], "-cookie") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing cookie" << endl;
                exit(1);
            }
            strcpy(cookie, argv[argindex]);
            cout << "Cookie of midi-out device set to " << cookie << endl;
            argindex++;
        }
        else if (strcmp(argv[argindex], "-incookie") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing cookie" << endl;
                exit(1);
            }
            strcpy(incookie, argv[argindex]);
            cout << "Cookie of midi-in device set to " << cookie << endl;
            argindex++;
        }
        else
        {
            cerr << "unknown option '" << argv[argindex] << "'" << endl;
            cerr << "possible options:" << endl;
            cerr << "    -trace" << endl;
            cerr << "    -port <portnr>" << endl;
            cerr << "    -host <hostname>" << endl;
            cerr << "    -cookie <midi-out device cookie>" << endl;
            cerr << "    -incookie <midi-in device cookie>" << endl;
            exit(1);
        }
    }

    cout << "Connecting to JamServer at " << host;
    cout << ", portnumber " << port << endl;

    pclient_ = new jamclient(cookie, incookie, host, port);
    if (!pclient_ -> connected())
    {
        cerr << "Couldn't connect to JamServer" << endl;
        exit(1);
    }
    octave_ = 4;
    volume_ = 127 - 32;
    instrument_ = 0;
}


int app::main()
{
    misc_handler_ = new c_misc_handler(this);
    char cmd[64];

    if (tk -> source("keyboard.tcl") != 0)
    {
        cerr << "Couldn't load 'keyboard.tcl'" << endl;
    }
    else
    {
        tk -> eval("keyboard::mk .k");  // create keyboard widget
        tk -> eval("pack .k -side top");
    }

    strcpy(cmd, cmd_keypress);
    strcat(cmd, " %A");
    tk -> root() -> bind("<KeyPress>", misc_handler_, cmd);

    strcpy(cmd, cmd_keyrelease);
    strcat(cmd, " %A");
    tk -> root() -> bind("<KeyRelease>", misc_handler_, cmd);

    label* l = new label(".l", "-relief sunken");
    l -> text("use keyboard to play");
    tk -> pack(l, "-side top");

    sinstrument_ = new scale(".s1", "-label instrument -orient horizontal "
                                    "-from 0 -to 127 -relief raised "
                                    "-showvalue 0");
    tk -> pack(sinstrument_, "-side top -fill both");
    sinstrument_ -> bind(misc_handler_, cmd_instrument);
    sinstrument_ -> set(0);

    svolume_= new scale(".s2", "-label volume -orient horizontal "
                               "-from 0 -to 127 -relief raised "
                               "-showvalue 0");
    tk -> pack(svolume_, "-side top -fill both");
    svolume_ -> bind(misc_handler_, cmd_volume);
    svolume_ -> set(100);

    soctave_ = new scale(".s3", "-label octave -orient horizontal "
                                "-from 0 -to 8 -relief raised "
                                "-showvalue 0 -length 200");
    tk -> pack(soctave_, "-side left -fill both");
    soctave_ -> bind(misc_handler_, cmd_octave);
    soctave_ -> set(4);

    button* b1 = new button(".b1", "-text \"Join jam\"");
    tk -> pack(b1, "-side left -fill both");
    b1 -> bind(misc_handler_, cmd_join);

    button* b2 = new button(".b2", "-text \"Leave jam\"");
    tk -> pack(b2, "-side left -fill both");
    b2 -> bind(misc_handler_, cmd_leave);

    button* bquit = new button(".bquit", "-text Quit");
    tk -> pack(bquit, "-side left -fill both -expand true");
    bquit -> bind(misc_handler_, cmd_quit);

// try to create Tcl commands 'play' and 'release'
    tk -> bind("play", misc_handler_);
    tk -> bind("release", misc_handler_);

// bind events for resetting and setting keyrepeat
    tk -> root() -> bind("<Motion>", misc_handler_, event_enter_window);
    tk -> root() -> bind("<Leave>", misc_handler_, event_leave_window);

// get original keyboard state
    display_ = Tk_Display(tk -> root() -> tkwin());
    XGetKeyboardControl(display_, &orgstate_);

    return OK;
}


void app::connect()
{
    cout << "join jam returns ";
    cout << pclient_ -> join() << endl;
// set initial instrument
    pclient_ -> program(instrument_);
}


void app::disconnect()
{
    pclient_ -> leave();
}
    

void app::keypress(char ch)
{
    int note = -1;
    int drum = -1;

    switch (ch)
    {
        case 'w':
            {
                char* users[midi_nr_channels];
                int i;

                for (i = 0; i < midi_nr_channels; i++)
                    users[i] = new char[64];

                if (!pclient_ -> users(users))
                {
                    cout << "Couldn't get users" << endl;
                }
                else
                {
                    cout << "Currently present Jammers:" << endl;
                    for (i = 0; i < midi_nr_channels; i++)
                    {
                        if (users[i][0] != '\0')
                        {
                            cout << "channel " << i << ": ";
                            cout << users[i] << endl;
                        }
                    }
                }

                for (i = 0; i < midi_nr_channels; i++)
                    delete[] users[i];
            }
            break;

        case '1':
            drum = midi_acoustic_bass_drum;
            break;
        case '2':
            drum = midi_acoustic_snare;
            break;
        case '3':
            drum = midi_closed_hihat;
            break;
        case '4':
            drum = midi_open_hihat;
            break;
        case '5':
            drum = midi_crash_cymbal;
            break;
        case '6':
            drum = midi_ride_cymbal;
            break;
        case '7':
            drum = midi_tambourine;
            break;
        case '8':
            drum = midi_hand_clap;
            break;
        case '9':
            drum = midi_low_tom;
            break;
        case '0':
            drum = midi_high_tom;
            break;
            
        case 'z':
            note = 0;
            break;
        case 's':
            note = 1;
            break;
        case 'x':
            note = 2;
            break;
        case 'd':
            note = 3;
            break;
        case 'c':
            note = 4;
            break;
        case 'v':
            note = 5;
            break;
        case 'g':
            note = 6;
            break;
        case 'b':
            note = 7;
            break;
        case 'h':
            note = 8;
            break;
        case 'n':
            note = 9;
            break;
        case 'j':
            note = 10;
            break;
        case 'm':
            note = 11;
            break;
        case ',':
            note = 12;
            break;
        case 'l':
            note = 13;
            break;
        case '.':
            note = 14;
            break;
        case ';':
            note = 15;
            break;
        case '/':
            note = 16;
            break;

        default:
            cout << "      'z'-'/' to play a note" << endl;
    }
    if (note >= 0)
    {
        if (!pclient_ -> note_on(note + octave_*12, volume_))
            cout << "couldn't send 'note on' to JamServer" << endl;
    }
    if (drum >= 0)
    {
        if (!pclient_ -> drum_on(drum, volume_))
            cout << "couldn't send 'drum on' to JamServer" << endl;
    }
}


void app::keyrelease(char ch)
{
    int note = -1;
    int drum = -1;

    switch (ch)
    {
        case 'z':
            note = 0;
            break;
        case 's':
            note = 1;
            break;
        case 'x':
            note = 2;
            break;
        case 'd':
            note = 3;
            break;
        case 'c':
            note = 4;
            break;
        case 'v':
            note = 5;
            break;
        case 'g':
            note = 6;
            break;
        case 'b':
            note = 7;
            break;
        case 'h':
            note = 8;
            break;
        case 'n':
            note = 9;
            break;
        case 'j':
            note = 10;
            break;
        case 'm':
            note = 11;
            break;
        case ',':
            note = 12;
            break;
        case 'l':
            note = 13;
            break;
        case '.':
            note = 14;
            break;
        case ';':
            note = 15;
            break;
        case '/':
            note = 16;
            break;

        case '1':
            drum = midi_acoustic_bass_drum;
            break;
        case '2':
            drum = midi_acoustic_snare;
            break;
        case '3':
            drum = midi_closed_hihat;
            break;
        case '4':
            drum = midi_open_hihat;
            break;
        case '5':
            drum = midi_crash_cymbal;
            break;
        case '6':
            drum = midi_ride_cymbal;
            break;
        case '7':
            drum = midi_tambourine;
            break;
        case '8':
            drum = midi_hand_clap;
            break;
        case '9':
            drum = midi_low_tom;
            break;
        case '0':
            drum = midi_high_tom;
            break;
    }
    if (note >= 0)
    {
        if (!pclient_ -> note_off(note + octave_*12, 127))
            cout << "couldn't send 'note off' to JamServer" << endl;
    }
    if (drum >= 0)
    {
        if (!pclient_ -> drum_off(drum, volume_))
            cout << "couldn't send 'drum off' to JamServer" << endl;
    }
}


void app::set_instrument()
{
    char buf[256];

    instrument_ = sinstrument_ -> get();
// tell JamServer new instrument
    pclient_ -> program(instrument_);
// change label of scale to show new instrument
    sprintf(buf, "-label \"instrument: %d, %s\"", instrument_,
            instrument_name[instrument_]);
    sinstrument_ -> configure(buf);
}


void app::set_volume()
{
    char buf[256];

    volume_ = svolume_ -> get();
// change label of scale to show new volume
    sprintf(buf, "-label \"volume: %d\"", volume_);
    svolume_ -> configure(buf);
}


void app::set_octave()
{
    char buf[256];

    octave_ = soctave_ -> get();
// change label of scale to show new volume
    sprintf(buf, "-label \"octave: %d\"", octave_);
    soctave_ -> configure(buf);
}


void app::keyboard_press(const char* playstr)
{
    int octave = 0;
    int note = 0;

    sscanf(playstr, "%d.%d", &octave, &note);
    octave = octave - 7 + octave_;

    if (!pclient_ -> note_on(note + octave*12, volume_))
        cout << "couldn't send 'note on' to JamServer" << endl;
}


void app::keyboard_release(const char* releasestr)
{
    int octave = 0;
    int note = 0;

    sscanf(releasestr, "%d.%d", &octave, &note);
    octave = octave - 7 + octave_;

    if (!pclient_ -> note_off(note + octave*12, volume_))
        cout << "couldn't send 'note off' to JamServer" << endl;
}


// to make sure we don't disrupt the 'normal' keyboard state, we turn
// auto-repeat off when _some_ window in the app is entered, and we turn it
// back on when the mouse-cursor leaves the app
void app::enter_window()
{
    XAutoRepeatOff(display_);
}


void app::leave_window()
{
    int winx, winy, height, width, mousex, mousey;

    sscanf(tk -> evaluate("winfo x ."), "%d", &winx);
    sscanf(tk -> evaluate("winfo y ."), "%d", &winy);
    sscanf(tk -> evaluate("winfo width ."), "%d", &width);
    sscanf(tk -> evaluate("winfo height ."), "%d", &height);
    sscanf(tk -> evaluate("winfo pointerx ."), "%d", &mousex);
    sscanf(tk -> evaluate("winfo pointery ."), "%d", &mousey);

    if ((mousex < winx) || (mousex >= winx + width) ||
        (mousey < winy) || (mousey >= winy + height))
    {
        if (orgstate_.global_auto_repeat)
            XAutoRepeatOn(display_);
    }
}


void app::quit()
{
// make sure global auto repeat is restored to the original state
    if (orgstate_.global_auto_repeat)
        XAutoRepeatOn(display_);
// wait for X to process the request
    tk -> update();

// not the nicest way to shut down, but it works...
    exit(0);
}

//----------------------------- main -----------------------

int main(int argc, char* argv[]) 
{
    cout << "HushJammer, version 0.5" << endl;

    app a(argc, argv);
    a.run();

    return 0;
}
