// jamserver.cc

#include <cs/dataconn.h>
#include <cs/tcp/server.h>
#include <cs/tcp/client.h>
#include <cs/samlib.h>
#include <genesis/msgparser.h>
#include <genesis/msgtypes.h>
#include <genesis/device.h>
#include <genesis/devices.h>
#include <genesis/midi.h>

#include <iostream.h>
#include <signal.h>
#include <stdlib.h>
#include <string.h>


const char version_str[] = "JamServer, version 0.03";

int user_interrupt = 0;

void inthandler(int /* signr */)
{
    user_interrupt = 1;
}

//-------------------- administration per jamclient -------------------

class clientdata
{
public:
    clientdata();
    clientdata(data_connection* conn, int id, int devout, int devin, 
               const char* username);
    ~clientdata();

    data_connection* connection();
    const char* username() { return username_; }
    int jam_id();
    int device_out();
    int device_in();

protected:
    data_connection* conn_;
    int jam_id_;
    int device_id_out_;
    int device_id_in_;
    char* username_;
};


clientdata::clientdata()
{
    conn_ = NULL;
    jam_id_ = -1;
    device_id_out_ = -1;
    device_id_in_ = -1;
    username_ = NULL;
}


clientdata::clientdata(data_connection* conn, int id, int devout, int devin, 
                       const char* username)
{
    conn_ = conn;
    jam_id_ = id;
    device_id_out_ = devout;
    device_id_in_ = devin;
    username_ = strdup(username);
}


clientdata::~clientdata()
{
    free(username_);
}


data_connection* clientdata::connection()
{
    return conn_;
}


int clientdata::jam_id()
{
    return jam_id_;
}


int clientdata::device_out()
{
    return device_id_out_;
}


int clientdata::device_in()
{
    return device_id_in_;
}

//---------------------------- jamserver definition ---------------------

int jam_trace = 0;
const int max_nr_jam_ids = midi_nr_channels;
const int jam_not_allocated = -1;
const int jam_no_device = 99999;

class jamserver: public tcp_server
{
public:
    jamserver(int jamport, const char* edenhost, int edenport);
    ~jamserver();

    int main();

protected:
    void process(data_connection* conn, const char* msg, int msglen);
    void receive_midi(data_connection* conn, const char* msg, int msglen);
    void remove_jam_client(data_connection* conn, const char* msg, int msglen);
        // gentle removal; client sent 'leave-jam' message
    void remove_jam_client(data_connection* conn);
        // rough removal; client crashed or just quit
    void add_jam_client(data_connection* conn, const char* msg, int msglen);
    void give_user_list(data_connection* conn, const char* msg, int msglen);
    void release_client_devices(int nrclient);
    void reject_jam(data_connection* conn);

    clientdata* clientdata_[max_nr_jam_ids];
        // index in clientdata_ is the jam_id of a client
    msgparser* msgp_;
    tcp_client* client_;
};

//------------------------ jamserver implementation ---------------------

jamserver::jamserver(int jamport, const char* edenhost, int edenport)
         :tcp_server(jamport)
{
    int i;

    for (i = 0; i < max_nr_jam_ids; i++)
        clientdata_[i] = NULL;

    msgp_ = new msgparser();
    client_ = new tcp_client(edenhost, edenport);
    if (!client_ -> connected())
        halt("Couldn't connect to Eden");
}


jamserver::~jamserver()
{
    delete msgp_;
}


int jamserver::main()
{
    char buffer[1024];
    int nrbytes = 0;
    int quit = 0;

    cerr << "Server is at port " << portnr() << endl;

    signal(SIGINT, inthandler);

    while((!user_interrupt) && (!quit))
    {
        if (!connected())
            return -1;

        fd_set edenfd;

        FD_ZERO(&edenfd);
        FD_SET(client_ -> fd(), &edenfd);

        data_connection* conn = select(&edenfd);
 
        if (conn != NULL)
        {
            nrbytes = readmsg(conn, buffer, sizeof(buffer));
            if (jam_trace)
                cout << "Received " << nrbytes << " bytes" << endl;
     
            if (nrbytes > 0)
                process(conn, buffer, nrbytes);
            else
            {
                if (jam_trace)
                    cout << "Probably disconnecting client or device" << endl;
                if ((conn == NULL) || (!conn -> connected()))
                {
                    if (jam_trace)
                        cout << "yep, something not connected anymore" << endl;
                    remove_jam_client(conn);
                    close(conn);
                }
                else
                {
                    if (jam_trace)
                        cout << "    nope, other error" << endl;
                }
            }
        }
        else if (FD_ISSET(client_ -> fd(), &edenfd))
        {
            if (!client_ -> ready())
                continue;           // probably interrupt

            nrbytes = client_ -> readmsg(buffer, sizeof(buffer));

            if (nrbytes <= 0)
            {
                cout << "server closed connection" << endl;
                quit = 1;
                continue;
            }

            if (msgp_ -> type(buffer) != msg_midi_out)
            {
                if (jam_trace)
                    cout << "unexpected message" << endl;
                continue;
            }

            receive_midi(NULL, buffer, nrbytes);
        }
        else
        {
            if (jam_trace)
                cout << "select returns NULL" << endl;
        }
    }
 
    return 0;
}


void jamserver::process(data_connection* conn, const char* msg, int msglen)
{
    if (jam_trace > 1)
        cout << "jamserver::process" << endl;
    switch (msgp_ -> type(msg))
    {
        case msg_join_jam:
            add_jam_client(conn, msg, msglen);
            break;
        case msg_leave_jam:
            remove_jam_client(conn, msg, msglen);
            break;
        case msg_midi_out:
            receive_midi(conn, msg, msglen);
            break;
        case msg_ask_users:
            give_user_list(conn, msg, msglen);
            break;
        default:
            if (jam_trace)
            {
                cout << "jamserver::process : other message (";
                cout << msgp_ -> type(msg) << ")" << endl;
            }
    }
}


void jamserver::add_jam_client(data_connection* conn, const char* msg,
                               int /* msglen */)
{
    char outmsg[64];
    int outmsglen;
    char outcookie[cookie_size+1];
    char incookie[cookie_size+1];
    char username[64];
    int i;
    int slot = -1;
    int outdevice = jam_no_device;
    int indevice = jam_no_device;

    if (jam_trace > 1)
        cout << "jamserver::add_jam_client" << endl;

// find free client-slot
    for (i = 0; i < max_nr_jam_ids; i++)
    {
    // don't allocate the percussion channel!
        if ((i != midi_percussion_channel) && (clientdata_[i] == NULL))
        {
            slot = i;
            break;
        }
    }
    if (slot < 0)
    {
    // no free slot
        reject_jam(conn);
        if (jam_trace)
            cout << "jamserver::add_jam_client : no free slot" << endl;
        return;
    }

    if (!msgp_ -> parse_join_jam(msg, outcookie, incookie, username))
    {
        warn("jamserver::add_jam_client : couldn't parse msg");
        return;
    }

    if (jam_trace)
    {
        cout << "jamserver::add_jam_client : outcookie = " << outcookie;
        cout << ", incookie = " << incookie << ", user " << username << endl;
    }

    if (outcookie[0] != 0)
    {
    // request midi-out device from Eden
        outmsglen = msgp_ -> create_request_dev(outmsg, dev_midi_out, 
                                                outcookie);
        client_ -> writemsg(outmsg, outmsglen);

    // 3 second time-out
        timeval tv = {3, 0};
        if (!client_ -> ready(&tv))
        {
            reject_jam(conn);
            return;
        }

        if (!client_ -> readmsg(outmsg, sizeof(outmsg)))
            halt("Connection to Eden lost");

        if (msgp_ -> type(outmsg) != msg_grant_dev)
        {
            reject_jam(conn);
            return;
        }
        outdevice = msgp_ -> device_id(outmsg);
    }

    if (incookie[0] != 0)
    {
    // request midi-in device from Eden
        outmsglen = msgp_ -> create_request_dev(outmsg, dev_midi_in, 
                                                incookie);
        client_ -> writemsg(outmsg, outmsglen);

    // 3 second time-out
        timeval tv = {3, 0};
        if (!client_ -> ready(&tv))
        {
            reject_jam(conn);
            return;
        }

        if (!client_ -> readmsg(outmsg, sizeof(outmsg)))
            halt("Connection to Eden lost");

        if (msgp_ -> type(outmsg) != msg_grant_dev)
        {
            reject_jam(conn);
            return;
        }
        indevice = msgp_ -> device_id(outmsg);
    }

    outmsglen = msgp_ -> create_accept_jam(outmsg, slot);
    if (!writemsg(conn, outmsg, outmsglen))
        warn("jamserver::add_jam_client : couldn't send accept msg");

    clientdata_[slot] = new clientdata(conn, slot, outdevice, indevice, 
                                       username);
}


void jamserver::remove_jam_client(data_connection* /* conn */, const char* msg,
                                  int /* msglen */)
{
    int jam_id;

    if (!msgp_ -> parse_leave_jam(msg, jam_id))
    {
        if (jam_trace)
            warn("jamserver::remove_jam_client : couldn't parse msg");
        return;
    }

    if ((jam_id < 0) || (jam_id >= max_nr_jam_ids) || 
        (clientdata_[jam_id] == NULL))
    {
        if (jam_trace)
            warn("jamserver::remove_jam_client : invalid id %d", jam_id);
        return;
    }

    release_client_devices(jam_id);
    delete clientdata_[jam_id];
    clientdata_[jam_id] = NULL;

// don't close connection, maybe jamclient wants to join again
}


void jamserver::remove_jam_client(data_connection* conn)
{
    int i;

    for (i = 0; i < max_nr_jam_ids; i++)
    {
        if ((clientdata_[i] != NULL) &&
            (clientdata_[i] -> connection() == conn))
        {
            release_client_devices(i);
            delete clientdata_[i];
            clientdata_[i] = NULL;
        // don't close connection, maybe jamclient wants to join again
        // yeah right, when he's quit or crashed...
        }
    }
}


void jamserver::release_client_devices(int nrclient)
{
    if (clientdata_[nrclient] -> device_out() != jam_no_device)
    {
    // release device
        char outmsg[64];
        int outmsglen = 0;

        if (jam_trace)
            cout << "deleting out-device of client " << nrclient << endl;

        outmsglen = msgp_ -> create_release_dev(outmsg, 
                                      clientdata_[nrclient] -> device_out());
        client_ -> writemsg(outmsg, outmsglen);
    }
    
    if (clientdata_[nrclient] -> device_in() != jam_no_device)
    {
    // release device
        char outmsg[64];
        int outmsglen = 0;

        if (jam_trace)
            cout << "deleting in-device of client " << nrclient << endl;

        outmsglen = msgp_ -> create_release_dev(outmsg, 
                                      clientdata_[nrclient] -> device_in());
        client_ -> writemsg(outmsg, outmsglen);
    }
}


void jamserver::receive_midi(data_connection* /* conn */, const char* msg,
                             int msglen)
{
    char mididata[1024];
    int midilen;
    char outmsg[1024];
    int outmsglen;
    int ids[max_nr_jam_ids];
    int index = 0;
    int i;

    midilen = msgp_ -> get_midi(msg, msglen, mididata);
    
// build id-list
    for (i = 0; i < max_nr_jam_ids; i++)
    {
        if ((clientdata_[i]  != NULL) && 
            (clientdata_[i] -> device_out() != jam_no_device))
        {
            ids[index] = clientdata_[i] -> device_out();
            index++;
        }
    }

    outmsglen = msgp_ -> create_midi_multicast(outmsg, 0, ids, index,
                                               mididata, midilen);
    if (!client_ -> writemsg(outmsg, outmsglen))
    {
        if (jam_trace)
            warn("jamserver::receive_midi : couldn't send midi multicast");
    }
}


void jamserver::give_user_list(data_connection* conn, const char* /*msg*/,
                               int /*msglen*/)
{
    char* users[midi_nr_channels];
    char outmsg[1024];
    int outmsglen;
    int i;

    for (i = 0; i < max_nr_jam_ids; i++)
    {
        if (clientdata_[i] != NULL)
            users[i] = strdup(clientdata_[i] -> username());
        else
            users[i] = NULL;
    }

    outmsglen = msgp_ -> create_give_users(outmsg, users);

    if (!writemsg(conn, outmsg, outmsglen))
    {
        if (jam_trace)
            warn("jamserver::give_user_list : couldn't send userlist");
    }

    for (i = 0; i < max_nr_jam_ids; i++)
    {
        if (users[i] != NULL)
            free(users[i]);
    }
}


void jamserver::reject_jam(data_connection* conn)
{
    char outmsg[64];
    char outmsglen;

    outmsglen = msgp_ -> create_reject_jam(outmsg);

    if (outmsglen <= 0)           // bad luck...
        return;

    if (!writemsg(conn, outmsg, outmsglen))
        warn("jamserver::reject_jam : couldn't send reject msg");
}

//------------------------------- main ------------------------------

int main(int argc, char* argv[])
{
    int argindex = 1;
    char edenhost[64] = EDEN_DEFAULT_HOST;
    int edenport = eden_default_port;
    int jamport = jamserver_default_port;

    cout << version_str << endl << endl;

    while (argindex < argc)
    {
        if (strcmp(argv[argindex], "-edenhost") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing hostname" << endl;
                exit(1);
            }
            strcpy(edenhost, argv[argindex]);
            cout << "EDEN-host set to " << edenhost << endl;
            argindex++;
        }
        else if (strcmp(argv[argindex], "-edenport") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing portnr" << endl;
                exit(1);
            }
            edenport = atoi(argv[argindex]);
            cout << "EDEN-port set to " << edenport << endl;
            argindex++;
        }
        else if (strcmp(argv[argindex], "-jamport") == 0)
        {
            argindex++;
            if (argindex >= argc)
            {
                cerr << "Missing portnr" << endl;
                exit(1);
            }
            jamport = atoi(argv[argindex]);
            cout << "JamServer-port set to " << jamport << endl;
            argindex++;
        }
        else if (strcmp(argv[argindex], "-trace") == 0)
        {
            jam_trace = 1;
            cout << "Tracing on" << endl;
            argindex++;
        }
        else
        {
            cerr << "Unknown option '" << argv[argindex] << "'" << endl;
            cerr << "Possible options:" << endl;
            cerr << "    -edenhost <hostname>" << endl;
            cerr << "    -edenport <portnr>" << endl;
            cerr << "    -jamport <portnr>" << endl;
            cerr << "    -trace" << endl;
            exit(1);
        }
    }

    jamserver* js = new jamserver(jamport, edenhost, edenport);
    js -> main();
    delete js;

    return 0;
}

