/**
Defines the class handler, which is
an ancestor of events as well as widgets
and items, as expressed in the 
See the
<a href=http://www.cs.vu.nl/~eliens/hush/structure.html#Widgets>class hierarchy</a>
See man:handler(4), <hush/hush.h> 
**/

#ifndef HANDLER_H
#define HANDLER_H

#include <hush/export.h>

#ifdef __GNUG__
#define NOHANDLER
#endif

//Unfortunately, OK for Tcl is 0,
//whereas OK for the simulation packages is 1. OK?


#undef OK
#define OK 0
#undef ERROR
#define ERROR 1

// TODO: is a handler or thread or what
// #include <hush/thread.h>

class event;
class kit;

class term;
class dvstring;
class iter_type;
class dictionary_type;
class dvlist_type;
class garbage_type;
class dvbinding;

class ostream;
class istream;

class HUSH_API handler {
static int _no;                    // to keep count of handlers
public:

enum  { implicit = 0, explic }; /// modes -- GNU does not allow explicit!


	handler();
	virtual ~handler();
	
	virtual char* name() const;    	/// get handler's command name
	virtual void rename(const char*); /// (re)set handler's command name
	
	virtual char* description() const; /// Explaining text
	
	virtual char* mimetype() const; /// when acting as a plugin
	
	
	virtual char* category() const; /// handler's kind 
	
	virtual char* info() const;     /// 

	// <b>dispatching</b>
	
	virtual event* dispatch(event* e = 0);
	virtual int operator()();

// io

virtual ostream& out(const char* opts=0) const;
virtual ostream& err(const char* opts=0) const;
virtual istream& in(const char* opts=0)  const;

// Default methods (originally introduced for canvas)

virtual void press( event&  ) { }
virtual void release( event&  ) { }
virtual void keypress( event&  ) { }
virtual void keyrelease( event&  ) { }
virtual void motion( event&  ) { }
virtual void enter( event&  ) { }
virtual void leave( event&  ) { }
virtual void other( event& );

// Access to the (local) kit:

virtual kit* thekit() const; /// OBSOLETE

// Modalities

virtual int mode() const;
virtual int mode(int);
virtual int mode(const char*);

// <b>dependent handlers</b>

virtual void dependent(handler* e); // to define dependent handlers 
virtual void remove(handler* e); // to remove dependent handlers
virtual int contains(handler* e); // to check for cycles
virtual void process(event* e = 0); // to activate dependent handlers

// <b>garbage collection</b>

virtual void _declare(const handler* h) const; /// for the future

void _cleanup(const char* opts = 0) const; // deletes garbage!

void _register(const handler* h) const; /// for collection
void _register(const term* p) const; /// idem
void _register(const dvstring* p) const; /// idem
void _register(const dvlist_type* p) const; /// idem
void _register(const iter_type* p) const; /// idem
void _register(const dictionary_type* p) const; /// idem
void _register(const garbage_type* p) const; /// idem
void _register(const char* p) const; /// idem
void _register(const dvbinding* p) const; /// idem

// Options and Resources 

virtual int _option(const char*);
virtual void _option(const char*,int);
     
virtual handler* _resource(const char*);
virtual void  _resource(const char*, handler*);

private:
    char* _name;

protected:
    event* _event;
    kit* tk;

    virtual void _write(const char* s, const char* msg = "") const;

public:
    virtual void _debug(const char* s, const char* msg = "") const;
    virtual void _warning(const char* s, const char* msg = "") const;
    virtual void* _error(const char* s, const char* msg = "") const; //0
    virtual void _obsolete(const char* s, const char* msg = "") const;

// And to create meta handlers
    
protected:    virtual void _reparent(handler* h); // to outer context
protected:    virtual handler* _redirect(handler* h); // to inner self

public:       virtual void _alias( handler* caller = 0 ); // for 'global' body

// To access the inner components of a compound widget
    
//public:       handler* self() const; // self in context or this
//protected:    virtual int exists() const;    // self() != this

// To support 'layered' handler components

private:
	handler* _this() const; // _context() or this
  	int _exists() const;    // is there a context?

protected:
  	handler* _body() const; // the implementation
	handler* _context() const; // non-virtual
	handler* _self() const;    // idem

public: virtual int invariant() const; /// to check the handler's state


private:
void* _state;

public: /// utilities

char* flatten(int argc, char* argv[]) const;
char* quote(int argc, char* argv[]) const;
char* gensym(const char* s=0) const;

virtual void obsolete(const char* msg = "HANDLER") const; // is obsolete 
virtual int verify() const;

public: /// plugin methods

virtual void* javaclass() const; // whatever that is (ae)

virtual void _initialize(); 
virtual void _shutdown();
 
virtual void _map(); // AE
virtual void _resize();
  
virtual void* _save();
   
virtual int _newstream(void* stream , int seekable );
virtual int _writeready(void* stream );
virtual int _write(void* stream, int offs , int len , void* );
    
virtual int _destroystream(void* stream, int reason);
virtual int _asfile(void* stream, const char*);
virtual int _printinfo(void* );

virtual void _request(void* stream, void* range);

public:
static int created;
static int deleted;

static void* table;
static void declare(const char*, void*, int flag = 0);
static handler* lookup(const char*);
static void destroy();

public:
static ostream* _out;
static ostream* _err;
static istream* _in;

static int initialized;
static int initialize();
};

#endif
