// A list with modalities and properties
// solidified by jrvosse

#ifndef _DVLIST_H
#define _DVLIST_H

#include <hush/export.h>

#include <hush/flags.h>  // to check for disabled lists
#ifndef NOLIST

#include <hush/baselist.h>
#include <hush/sublist.h>
#include <hush/garbage.h>
#include <hush/iter.h>
#include <hush/assert.h>

template<class T>
class dvlist : public dvlist_type {
public:
  dvlist(const T* p = 0); 		/// ctor
  dvlist(const dvlist<T>& x);		/// shallow copy ctor
  dvlist<T>* clone() const;		/// explicit cloning
  dvlist<T>& operator=(const dvlist<T>& x); /// shallow assignment
  
  virtual ~dvlist();			/// dtor
  void destroy(); 			/// deletes all elements, not list
  
  void insert(const T* el, int before_current=0);/// insert
  void insert(const dvlist<T>& l);
  
  T* remove(const T* el = 0) ;		/// remove
  void remove(const dvlist<T>& l) ;		
  void remove(int index, int len=1) ;
  
  //// Alternative names 
  void put(const T* el); 		/// insert
  T* get();				/// remove
  
  int empty() const;			/// info
  int length() const; 
  int index() const; 			/// index of  current
  int contains(const T* el) const ;
  int count(const T* el = 0) const ;
  
  T* head() const; 			/// access
  T* current() const;
  T* last() const;

  T* operator[](int index);		/// current(i)
  T* operator[](const T* p);		/// find(p)

  dvlist<T>* tail() const;		/// return copy of list - head
  dvlist<T>* range(int from, int len) const; /// return copy of a range

  int operator!=(const dvlist<T>& x);	/// comparing
  int operator==(const dvlist<T>& x); 
  int operator<=(const dvlist<T>& x);
  int operator<(const dvlist<T>& x);
  
  T* find(const T* el);  		/// set cursor to first el found
  T* next(const T* el);  		/// set cursor to next elem found
  T* operator++(int);			/// postfix; dummy argument ignored
  T* operator--(int);			/// postfix; dummy argument ignored
  
  //// Modalities
  enum { _plain = 0, _ordered, _lifo, _fifo };  
  int mode() const; 
  int mode(int i);
  
  int plain()   const;
  int ordered() const;
  int lifo()    const;
  int fifo()    const;
  
  //// Modifications and replacement
  T* replace(const T* el);
  T* replace(const T* el, int pos, int len=1); 
  void rotate(int i=1);
  
  //// Abbreviations
  dvlist<T>& operator<<(const T* el); 		/// insert el
  dvlist<T>& operator<<(const dvlist<T>& l) ;	/// insert list
  
  dvlist<T>& operator>>(T*&  el); 		/// remove current
  dvlist<T>& operator<<(int i); 		/// rotate(i);
  dvlist<T>& operator>>(int i); 		/// rotate(-i);
  
  //// Iterator
  operator iter<T>&() const { 
	return * (iter<T>*) _body->iterator();
  }
  iter<T>* walk() {
  	return (iter<T>*) _body->iterator();
	}
  
  T* current(int i);				/// goto index i
  T* first();					/// goto index 0
  T* end();					/// goto index -1
  T* next();					/// index++
  T* previous();				/// index--
  
  //// Printing
  typedef ostream& (*format_type)(ostream&, const T*); /// for properties
  friend ostream& operator<<(ostream& os, const dvlist<T>& l) { l.print(os); return os; }
  static ostream& default_format(ostream& os, const T* e) { return os << e << endl; }
  
  void print(ostream& os) const;
  
  format_type format() const { return (format_type) _format; }
  void format(format_type fun) { _format = (format_type) fun; }
  
  //// Equality and order
  typedef int (*lcompare_type) (const dvlist<T>&, const dvlist<T>&);
  typedef int (*compare_type)  (const T*, const T*);
  typedef int (*predicate_type)(const T*, const dvlist<T>&);
  typedef T*  (*element_type)  (const T*, const dvlist<T>&);
  
  compare_type equality() const { return (compare_type) _equals; }
  void equality(const compare_type fun) { _equals = (compare_type) fun; }
  
  compare_type order() const { return (compare_type) _less; }
  void order(const compare_type fun) { _less = (compare_type) fun; }
  
  lcompare_type lequality() const { return (lcompare_type) _lequals; }
  void lequality(const lcompare_type fun) { _lequals = (lcompare_type) fun; }
  
  lcompare_type lorder() const { return (lcompare_type) _lless; }
  void lorder(const lcompare_type fun) { _lless = (lcompare_type) fun; }
  
  element_type insertion() const { return (element_type) _insert; }
  void insertion(const element_type fun) { _insert = (element_type) fun; }
  
  element_type extraction() const { return (element_type) _extract; }
  void extraction(const element_type fun) { _extract = (element_type) fun; }
  
  element_type selection() const { return (element_type) _select; }
  void selection(const element_type fun) { _select = (element_type) fun; }
  
  predicate_type predicate() const { return (predicate_type) _predicate; }
  void predicate(const predicate_type fun) { _predicate = (predicate_type) fun; }
  
  // dvlist<T>& select(const T* el) const; // return list with all elements equal to el
  
  int process(const predicate_type fun=0) const {
    predicate_type f = fun?fun:predicate();
    return _body->process( (void*) f ); 
  }

  dvlist<T>& apply(element_type fun=0) const {
  element_type f = fun?fun:selection();
  dvlist<T>* tmp = (dvlist<T>*) _body->apply( (void*) f );
  _register( tmp );
  tmp->_copy(*this);
  return *tmp;
  }

  #ifdef recursive_templates
  dvlist< dvlist<T> >& map( dvlist<T>* fun(T*, dvlist<T>*) ) {
    dvlist< dvlist<T> >*tmp = 
      (dvlist< dvlist<T> >*) _body->map( fun );
    return *tmp;
  }
  #endif // no_recursive_templates

  //// Default properties
  
  static int anything(const T*, const dvlist<T>&)  { return 1; }
  static T* identity(const T* x, const dvlist<T>&) { return (T*) x; }
  static int less(const T*, const T*) { return 0; } 
  static int equals(const T* e1, const T* e2) { return e1 == e2; } /// pointer equality!

  static int lequals(const dvlist<T>& l1, const dvlist<T>& l2) { 
    if ( l1.length() == l2.length() ) {
      iter<T> it1 = l1; iter<T> it2 = l2;
      T* p1 = 0; T* p2 = 0;
      while ( (p1 = it1()) && (p2 = it2()) ) {
        if ( ! l1.equality()(p1,p2) ) return 0;
      }
      return 1;
    }
    return 0;
  }

  static int lless(const dvlist<T>& l1, const dvlist<T>& l2) { 
    if ( l1.length() <= l2.length() ) {
      iter<T>& it1 = l1; iter<T>& it2 = l2;
      T* p1 = 0; T* p2 = 0;
      while ( (p1 = it1()) && (p2 = it2()) ) {
        if ( ! (l1.equality()(p1,p2) || l1.order()(p1,p2)  ) )
	  return 0;
      }
      return 1;
    }
    return 0;
  }

  //// System level usage only:
  
  friend class dvsublist<T>;			/// allows for list(i,j) = x
  dvsublist<T>& operator()(int from, int len);	/// return range

  friend class dvlistelement<T>;		/// allows for list(i) = x
  dvlistelement<T>& operator()();		/// current()
  dvlistelement<T>& operator()(int index);	/// current(i)
  dvlistelement<T>& operator()(const T* p);	/// find(p)
  
  void _copy( const dvlist<T>& p ); /// to transfer properties, functions
  void _destroy();		    /// destroys cells _and_ elements
  
  //// Registration of garbage
  
  void _register(const garbage_type* p) const;
  void _register(const dvlist<T>* p) const;
  void _register(const T* p) const;
  void _register(const dvsublist<T>* p) const;
  void _register(const dvlistelement<T>* p) const;
  
protected:
  dvbaselist* _body;
private:
  dvgarbage<T>* _garb;
  int _mode; /// modality
  void init();
  format_type _format;
  element_type _insert;
  element_type _extract;
  element_type _select;
  compare_type _equals;
  compare_type _less;
  lcompare_type _lequals;
  lcompare_type _lless;
  predicate_type _predicate;

public:
static void declare(const char* s, void* l, int flag = 0) {
	dvlist_type::declare(s,l,flag);
	}
static dvlist<T>* lookup(const char* s) {
	return (dvlist<T>*) dvlist_type::lookup(s);
	}
};

//// Special lists

template<class T>
class dvslist : public dvlist<T> {	// a list with semantic equality
public:
  dvslist() { equality( dvslist<T>::equals ); format( dvslist<T>::sformat ); }
  dvslist(const dvslist<T>& l): dvlist<T>(l) { }
  static int equals(const T* e1, const T* e2) { return *e1 == *e2; }
  static ostream& sformat(ostream& os, const T* e) {
    os << *e << endl; return os;
  }
};


template<class T>
class dvolist : public dvslist<T> {	// an slist with (semantic) ordening
public:
  dvolist() { mode( _ordered ); order( dvolist<T>::less ); }
  dvolist(const dvolist<T>& l): dvslist<T>(l) { }
  static int less(const T* e1, const T* e2) { return *e1 < *e2; }
};

//<ignore>

//// Implementation

template<class T> inline
dvlist<T>::dvlist(const T* p) {		/// ctor
  _garb = 0;  
  _body = new dvbaselist(this); 
  init(); 
  if (p) insert(p);
}

template<class T> inline
void dvlist<T>::init() {
  mode( _plain );
  insertion( identity );
  extraction( identity );
  selection( identity );
  predicate( anything );
  format( default_format );
  equality( equals );
  order( less );
  lequality( lequals );
  lorder( lless );
}

template<class T> inline
void dvlist<T>::_copy( const dvlist<T>& x) {
  mode( x.mode() );
  insertion( x.insertion() );
  extraction( x.extraction() );
  selection( x.selection() );
  predicate( x.predicate() );
  format( x.format() );
  equality( x.equality() );
  order( x.order() );
  lequality( x.lequality() );
  lorder( x.lorder() );
}

template<class T> inline
dvlist<T>::dvlist(const dvlist<T>& x) {		/// copy ctor
  _garb = 0;
  _body = x._body;			/// shallow copy!
  _copy( x );				/// copy properties
  _body->refs++; 			/// reference count
}

template<class T> inline
dvlist<T>::~dvlist() {			/// dtor
  if (_body->refs > 0) _body->refs--;
  else { destroy(); delete _body; }
  delete _garb; _garb = 0;
}

template<class T> inline
void dvlist<T>::destroy() { 	/// deletes all nodes, not elements 
  while (remove());
}

template<class T> inline
void dvlist<T>::_destroy() { 	/// deletes all nodes and elements 
  T* el;
  while (( el = remove() )) delete el;
}

template<class T> inline
dvlist<T>* dvlist<T>::clone() const {
  dvlist<T>* tmp = (dvlist<T>*) _body->clone();
  //tmp->_copy( *this );	/// Bug ??? AE: no, I dont think so
  return tmp; // no registration and, hence, no deletion. 
}

template<class T> inline
dvlist<T>& dvlist<T>::operator=(const dvlist<T>& x) { /// assignment
  if ( _body->refs == 0 ) { destroy(); delete _body; _body=0; }
  else _body->refs--; 
  _body = x._body; _body->refs++;	// shallow copy!
  _copy(x); /// take the properties of x
  return (*this);
}

template<class T> inline
int dvlist<T>::operator!=(const dvlist<T>& x) { return ! lequality()(*this,x); }

template<class T> inline
int dvlist<T>::operator==(const dvlist<T>& x) { return lequality()(*this,x); }

template<class T> inline
int dvlist<T>::operator<(const dvlist<T>& x)  { return lorder()(*this,x); }

template<class T> inline
int dvlist<T>::operator<=(const dvlist<T>& x) { return (*this==x) || (*this<x); }

template<class T> inline
int dvlist<T>::empty() const  { return _body->empty(); }

template<class T> inline
int dvlist<T>::index() const { return _body->index(); }

template<class T> inline
int dvlist<T>::length() const { return _body->length(); }

template<class T> inline
T* dvlist<T>::operator[](int index)  { return current(index); }

template<class T> inline
T* dvlist<T>::operator[](const T* p) { return find(p); }
template<class T> inline
dvlistelement<T>& dvlist<T>::operator()() { 
	require(current());
	dvlistelement<T> *tmp = new dvlistelement<T>(this);
	_register(tmp);
	return *tmp;
}

template<class T> inline
dvlistelement<T>& dvlist<T>::operator()(int index) { 
	T* c = current(index); require(c);
	return this->operator()(); 
}

template<class T> inline
dvlistelement<T>& dvlist<T>::operator()(const T* p) { 
	T* c = find(p); require(c);
	return this->operator()(); 
}

template<class T> inline
T* dvlist<T>::find(const T* el) {  	/// sets cursor
  return (T*) _body->find(el);
}

template<class T> inline
T* dvlist<T>::operator++(int) {				/// dummy argument ignored
  T* tmp = current();
  next();
  return tmp;
}

template<class T> inline
T* dvlist<T>::operator--(int) {				 /// dummy argument ignored
  T* tmp = current();
  previous();
  return tmp;
}

template<class T> inline
int dvlist<T>::contains(const T* el) const {
  return  _body->contains(el);
}

template<class T> inline
int dvlist<T>::count(const T* el) const {
  if (!el) return _body->length();
  else return _body->count(el);
}

template<class T> inline int dvlist<T>::mode() const { return _mode; }

template<class T> inline int dvlist<T>::mode(int i) { int tmp = _mode; _mode = i; return tmp; }

template<class T> inline int dvlist<T>::plain()   const { return mode() == _plain; }

template<class T> inline int dvlist<T>::ordered() const { return mode() == _ordered; }

template<class T> inline int dvlist<T>::lifo()    const { return mode() == _lifo; }

template<class T> inline int dvlist<T>::fifo()    const { return mode() == _fifo; }

//// Insertion and removal -- plain

template<class T> inline
void dvlist<T>::insert(const T* el, int before) { _body->insert((void*) el, before); }

template<class T> inline
void dvlist<T>::insert(const dvlist<T>& l) { _body->linsert(l); }

template<class T> inline
T* dvlist<T>::remove(const T* el) {
  if (!el) {
    if (lifo()) end();			/// set current to last
    if (fifo()) first();		/// set current to first
    return (T*) _body->remove(); 	/// remove current
  } else if (find(el))			/// set current to el
    return (T*) _body->remove(); 	/// remove current
  else return 0;
}

template<class T> inline
void dvlist<T>::remove(const dvlist<T>& l) {		
  _body->lremove((dvlist_type&) l);
}

template<class T> inline
void dvlist<T>::remove(int index, int len) {
  _body->remove(index, len);
}

template<class T> inline
void dvlist<T>::put(const T* el)  { insert(el); }

template<class T> inline
T*   dvlist<T>::get() 		 { return remove(); } 

//// Modifications and replacement

template<class T> inline
T* dvlist<T>::replace(const T* el) {
  return (T*) _body->replace(el);
}

template<class T> inline
T* dvlist<T>::replace(const T* el, int pos, int len) { 
  return (T*) _body->replace((void*) el, pos,len);
}

template<class T> inline
void dvlist<T>::rotate(int i) { _body->rotate(i); }

//// Abbreviations

template<class T> inline
dvlist<T>& dvlist<T>::operator<<(const T* e) { insert(e); return *this; }

template<class T> inline
dvlist<T>& dvlist<T>::operator<<(const dvlist<T>& l) {
  T* p = 0; iter<T>& it = l; while((p=it())) insert(p);
  return *this; 
}

template<class T> inline
dvlist<T>& dvlist<T>::operator>>(T*&  e) { e = remove(); return *this; }

template<class T> inline
dvlist<T>& dvlist<T>::operator<<(int i) { rotate(i);return *this; }

template<class T> inline
dvlist<T>& dvlist<T>::operator>>(int i) { rotate(-i);return *this; }

//// Access
template<class T> inline
T* dvlist<T>::head() const    { return (T*) _body->head(); }

template<class T> inline
T* dvlist<T>::current() const { return (T*) _body->current(); }

template<class T> inline
T* dvlist<T>::last() const    { return (T*) _body->last(); }

template<class T> inline
dvlist<T>* dvlist<T>::tail() const { return range(1,length()-1); }

template<class T> inline
dvlist<T>* dvlist<T>::range(int from, int len) const { 
	dvlist<T>* result = new dvlist<T>;
	dvlist<T>* l = (dvlist<T>*) this; /// cast const away
        int old_index = l->index();       /// save index;
 
        int leng = l->length();           /// normalize first
        int first = from % leng;          /// so that 0 <= first < length
        if (first<0) first += leng;
 
        result->_copy(*l);                /// copy properties
        for(int i=first; i< first+len && i<leng; i++)
                result->insert( l->current(i) );
        l->current(old_index);          /// restore index
	return result;
}

template<class T> inline
dvsublist<T>& dvlist<T>::operator()(int from, int len) { 
	dvsublist<T>* tmp = new dvsublist<T>(this, from, len);
	_register(tmp);
	return *tmp;
}

//// Iterator
////
//// template<class T> inline
//// dvlist<T>::operator iter<T>&() const { 
////  return * (iter<T>*) _body->iterator();
//// }

template<class T> inline
T* dvlist<T>::current(int i) { return (T*) _body->current(i); }

template<class T> inline
T* dvlist<T>::first()    { return (T*) _body->first(); }

template<class T> inline
T* dvlist<T>::end()     { return (T*) _body->end(); }

template<class T> inline
T* dvlist<T>::next()     { return (T*) _body->next(); }

template<class T> inline
T* dvlist<T>::previous() { return (T*) _body->previous(); }

template<class T> inline
T* dvlist<T>::next(const T* el) {  /// sets cursor
  return (T*) _body->next(el);
}

template<class T> inline
void dvlist<T>::print(ostream& os) const { _body->print(os); }

//// Selection

// template<class T> inline
// dvlist<T>& dvlist<T>::select(const T*) const { 
// 	return body->select(); 
// }

//// Registration of garbage

template<class T> inline
void dvlist<T>::_register(const garbage_type* p) const { _body->_register( (void*) p); }
template<class T> inline
void dvlist<T>::_register(const dvlist<T>* p) const { _body->l_register((void*)p); }
template<class T> inline
void dvlist<T>::_register(const dvsublist<T>* p) const { _body->s_register((void*)p); }
template<class T> inline
void dvlist<T>::_register(const dvlistelement<T>* p) const { _body->e_register((void*)p); }
template<class T> inline
void dvlist<T>::_register(const T* p) const     { 
  if (!_garb) ((dvlist<T>*)this)->_garb = new dvgarbage<T>;
  _garb->_register((T*)p);
}

//// Defines

#ifndef list
#define list dvlist
#endif

#ifndef slist
#define slist dvslist
#endif

#ifndef olist
#define olist dvolist
#endif

#endif // NOLIST

#endif
